@@ -25,6 +25,8 @@ LIBRARY_DIRNAMES = {
     'python-dotenv': 'dotenv',
     'pip-tools': 'piptools',
     'setuptools': 'pkg_resources',
+    'msgpack-python': 'msgpack',
+    'attrs': 'attr',
 }
 
 # from time to time, remove the no longer needed ones
@@ -43,6 +45,9 @@ HARDCODED_LICENSE_URLS = {
     'pytoml': 'https://github.com/avakar/pytoml/raw/master/LICENSE',
     'webencodings': 'https://github.com/SimonSapin/python-webencodings/raw/'
                     'master/LICENSE',
+    'requirementslib': 'https://github.com/techalchemy/requirementslib/raw/master/LICENSE',
+    'distlib': 'https://github.com/vsajip/distlib/raw/master/LICENSE.txt',
+    'pythonfinder': 'https://raw.githubusercontent.com/techalchemy/pythonfinder/master/LICENSE.txt'
 }
 
 FILE_WHITE_LIST = (
@@ -58,14 +63,14 @@ FILE_WHITE_LIST = (
     'vendor_pip.txt',
 )
 
-LIBRARY_RENAMES = {
-    'pip': 'pip9'
-}
-
 PATCHED_RENAMES = {
     'pip': 'notpip'
 }
 
+LIBRARY_RENAMES = {
+    'pip': 'pipenv.patched.notpip'
+}
+
 
 def drop_dir(path):
     if path.exists() and path.is_dir():
@@ -84,14 +89,16 @@ def log(msg):
     print('[vendoring.%s] %s' % (TASK_NAME, msg))
 
 
+def _get_git_root(ctx):
+    return Path(ctx.run('git rev-parse --show-toplevel', hide=True).stdout.strip())
+
+
 def _get_vendor_dir(ctx):
-    git_root = ctx.run('git rev-parse --show-toplevel', hide=True).stdout
-    return Path(git_root.strip()) / 'pipenv' / 'vendor'
+    return _get_git_root(ctx) / 'pipenv' / 'vendor'
 
 
 def _get_patched_dir(ctx):
-    git_root = ctx.run('git rev-parse --show-toplevel', hide=True).stdout
-    return Path(git_root.strip()) / 'pipenv' / 'patched'
+    return _get_git_root(ctx) / 'pipenv' / 'patched'
 
 
 def clean_vendor(ctx, vendor_dir):
@@ -133,8 +140,12 @@ def rewrite_imports(package_dir, vendored_libs, vendor_dir):
 
 def rewrite_file_imports(item, vendored_libs, vendor_dir):
     """Rewrite 'import xxx' and 'from xxx import' for vendored_libs"""
-    text = item.read_text(encoding='utf-8')
-    renames = PATCHED_RENAMES if vendor_dir.name == 'patched' else LIBRARY_RENAMES
+    log('Reading file: %s' % item)
+    try:        
+        text = item.read_text(encoding='utf-8')
+    except UnicodeDecodeError:
+        text = item.read_text(encoding='cp1252')
+    renames = LIBRARY_RENAMES
     for k in LIBRARY_RENAMES.keys():
         if k not in vendored_libs:
             vendored_libs.append(k)
@@ -162,7 +173,7 @@ def rewrite_file_imports(item, vendored_libs, vendor_dir):
 
 def apply_patch(ctx, patch_file_path):
     log('Applying patch %s' % patch_file_path.name)
-    ctx.run('git apply --verbose %s' % patch_file_path)
+    ctx.run('git apply --ignore-whitespace --verbose %s' % patch_file_path)
 
 
 @invoke.task
@@ -292,7 +303,8 @@ def vendor(ctx, vendor_dir, rewrite=True):
     drop_dir(vendor_dir / 'tests')
 
     # Detect the vendored packages/modules
-    vendored_libs = detect_vendored_libs(vendor_dir)
+    vendored_libs = detect_vendored_libs(_get_vendor_dir(ctx))
+    patched_libs = detect_vendored_libs(_get_patched_dir(ctx))
     log("Detected vendored libraries: %s" % ", ".join(vendored_libs))
 
     # Apply pre-patches
@@ -303,15 +315,13 @@ def vendor(ctx, vendor_dir, rewrite=True):
             if not patch.name.startswith('_post'):
                 apply_patch(ctx, patch)
 
+    log("Removing scandir library files...")
+    remove_all(vendor_dir.glob('*.so'))
+
     # Global import rewrites
     log('Renaming specified libs...')
     for item in vendor_dir.iterdir():
         if item.is_dir():
-            if item.name == 'requests' and not (item / 'cacert.pem').exists():
-                if 'certifi' in vendored_libs:
-                    cert = vendor_dir / 'certifi' / 'cacert.pem'
-                    copy_to = item / 'cacert.pem'
-                    copy_to.write_bytes(cert.read_bytes())
             if rewrite:
                 log('Rewriting imports for %s...' % item)
                 rewrite_imports(item, vendored_libs, vendor_dir)
@@ -321,13 +331,31 @@ def vendor(ctx, vendor_dir, rewrite=True):
                 rewrite_file_imports(item, vendored_libs, vendor_dir)
     write_backport_imports(ctx, vendor_dir)
     log('Applying post-patches...')
-    patches = patch_dir.glob('*.patch' if not is_patched else '_post*.patch') 
+    patches = patch_dir.glob('*.patch' if not is_patched else '_post*.patch')
     for patch in patches:
         apply_patch(ctx, patch)
     if is_patched:
         piptools_vendor = vendor_dir / 'piptools' / '_vendored'
         if piptools_vendor.exists():
             drop_dir(piptools_vendor)
+        msgpack = vendor_dir / 'notpip' / '_vendor' / 'msgpack'
+        if msgpack.exists():
+            remove_all(msgpack.glob('*.so'))
+
+
+@invoke.task
+def redo_imports(ctx, library):
+    vendor_dir = _get_vendor_dir(ctx)
+    log('Using vendor dir: %s' % vendor_dir)
+    vendored_libs = detect_vendored_libs(vendor_dir)
+    item = vendor_dir / library
+    library_name = vendor_dir / '{0}.py'.format(library)
+    log("Detected vendored libraries: %s" % ", ".join(vendored_libs))
+    log('Rewriting imports for %s...' % item)
+    if item.is_dir():
+        rewrite_imports(item, vendored_libs, vendor_dir)
+    else:
+        rewrite_file_imports(library_name, vendored_libs, vendor_dir)
 
 
 @invoke.task
@@ -457,6 +485,25 @@ def extract_license_member(vendor_dir, tar, member, name):
         dest.write_bytes(tar.read(member))
 
 
+@invoke.task()
+def generate_patch(ctx, package_path, patch_description, base='HEAD'):
+    pkg = Path(package_path)
+    if len(pkg.parts) != 2 or pkg.parts[0] not in ('vendor', 'patched'):
+        raise ValueError('example usage: generate-patch patched/pew some-description')
+    if patch_description:
+        patch_fn = '{0}-{1}.patch'.format(pkg.parts[1], patch_description)
+    else:
+        patch_fn = '{0}.patch'.format(pkg.parts[1])
+    command = 'git diff {base} -p {root} > {out}'.format(
+        base=base,
+        root=Path('pipenv').joinpath(pkg),
+        out=Path(__file__).parent.joinpath('patches', pkg.parts[0], patch_fn),
+    )
+    with ctx.cd(str(_get_git_root(ctx))):
+        log(command)
+        ctx.run(command)
+
+
 @invoke.task(name=TASK_NAME)
 def main(ctx):
     vendor_dir = _get_vendor_dir(ctx)
@@ -465,10 +512,10 @@ def main(ctx):
     clean_vendor(ctx, vendor_dir)
     clean_vendor(ctx, patched_dir)
     vendor(ctx, vendor_dir)
-    vendor(ctx, patched_dir, rewrite=False)
+    vendor(ctx, patched_dir, rewrite=True)
     download_licenses(ctx, vendor_dir)
     download_licenses(ctx, patched_dir, 'patched.txt')
-    for pip_dir in [vendor_dir / 'pip9', patched_dir / 'notpip']:
+    for pip_dir in [patched_dir / 'notpip']:
         _vendor_dir = pip_dir / '_vendor'
         vendor_src_file = vendor_dir / 'vendor_pip.txt'
         vendor_file = _vendor_dir / 'vendor.txt'
